@@ -1,5 +1,5 @@
 # -*- test-case-name: twisted.web.test.test_webclient -*-
-# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2009 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
 """
@@ -9,43 +9,52 @@ HTTP client.
 import os, types
 from urlparse import urlunparse
 
+from twisted.python import log
 from twisted.web import http
 from twisted.internet import defer, protocol, reactor
 from twisted.python import failure
 from twisted.python.util import InsensitiveDict
 from twisted.web import error
+from twisted.web.http_headers import Headers
+from twisted.python.compat import set
 
 
 class PartialDownloadError(error.Error):
-    """Page was only partially downloaded, we got disconnected in middle.
+    """
+    Page was only partially downloaded, we got disconnected in middle.
 
-    The bit that was downloaded is in the response attribute.
+    @ivar response: All of the response body which was downloaded.
     """
 
 
 class HTTPPageGetter(http.HTTPClient):
 
     quietLoss = 0
-    followRedirect = 1
+    followRedirect = True
     failed = 0
 
+    _specialHeaders = set(('host', 'user-agent', 'cookie', 'content-length'))
+
     def connectionMade(self):
         method = getattr(self.factory, 'method', 'GET')
         self.sendCommand(method, self.factory.path)
         self.sendHeader('Host', self.factory.headers.get("host", self.factory.host))
         self.sendHeader('User-Agent', self.factory.agent)
-        if self.factory.cookies:
-            l=[]
-            for cookie, cookval in self.factory.cookies.items():
-                l.append('%s=%s' % (cookie, cookval))
-            self.sendHeader('Cookie', '; '.join(l))
         data = getattr(self.factory, 'postdata', None)
         if data is not None:
             self.sendHeader("Content-Length", str(len(data)))
+
+        cookieData = []
         for (key, value) in self.factory.headers.items():
-            if key.lower() != "content-length":
+            if key.lower() not in self._specialHeaders:
                 # we calculated it on our own
                 self.sendHeader(key, value)
+            if key.lower() == 'cookie':
+                cookieData.append(value)
+        for cookie, cookval in self.factory.cookies.items():
+            cookieData.append('%s=%s' % (cookie, cookval))
+        if cookieData:
+            self.sendHeader('Cookie', '; '.join(cookieData))
         self.endHeaders()
         self.headers = {}
 
@@ -84,6 +93,18 @@ class HTTPPageGetter(http.HTTPClient):
         if self.followRedirect:
             scheme, host, port, path = \
                 _parse(url, defaultPort=self.transport.getPeer().port)
+
+            self.factory._redirectCount += 1
+            if self.factory._redirectCount >= self.factory.redirectLimit:
+                err = error.InfiniteRedirection(
+                    self.status,
+                    'Infinite redirection detected',
+                    location=url)
+                self.factory.noPage(failure.Failure(err))
+                self.quietLoss = True
+                self.transport.loseConnection()
+                return
+
             self.factory.setURL(url)
 
             if self.factory.scheme == 'https':
@@ -100,10 +121,14 @@ class HTTPPageGetter(http.HTTPClient):
                 failure.Failure(
                     error.PageRedirect(
                         self.status, self.message, location = url)))
-        self.quietLoss = 1
+        self.quietLoss = True
         self.transport.loseConnection()
 
-    handleStatus_302 = lambda self: self.handleStatus_301()
+    def handleStatus_302(self):
+        if self.afterFoundGet:
+            self.handleStatus_303()
+        self.handleStatus_301()
+
 
     def handleStatus_303(self):
         self.factory.method = 'GET'
@@ -159,6 +184,11 @@ class HTTPPageDownloader(HTTPPageGetter):
             self.factory.pagePart(data)
 
     def handleResponseEnd(self):
+        if self.length:
+            self.transmittingPage = 0
+            self.factory.noPage(
+                failure.Failure(
+                    PartialDownloadError(None)))
         if self.transmittingPage:
             self.factory.pageEnd()
             self.transmittingPage = 0
@@ -190,6 +220,24 @@ class HTTPClientFactory(protocol.ClientFactory):
     @type response_headers: dict
     @ivar response_headers: The headers that were specified in the
           response from the server.
+
+    @type method: str
+    @ivar method: The HTTP method to use in the request.  This should be one of
+        OPTIONS, GET, HEAD, POST, PUT, DELETE, TRACE, or CONNECT (case
+        matters).  Other values may be specified if the server being contacted
+        supports them.
+
+    @type redirectLimit: int
+    @ivar redirectLimit: The maximum number of HTTP redirects that can occur
+          before it is assumed that the redirection is endless.
+
+    @type afterFoundGet: C{bool}
+    @ivar afterFoundGet: Deviate from the HTTP 1.1 RFC by handling redirects
+        the same way as most web browsers; if the request method is POST and a
+        302 status is encountered, the redirect is followed with a GET method
+
+    @type _redirectCount: int
+    @ivar _redirectCount: The current number of HTTP redirects encountered.
     """
 
     protocol = HTTPPageGetter
@@ -202,11 +250,14 @@ class HTTPClientFactory(protocol.ClientFactory):
 
     def __init__(self, url, method='GET', postdata=None, headers=None,
                  agent="Twisted PageGetter", timeout=0, cookies=None,
-                 followRedirect=1):
-        self.protocol.followRedirect = followRedirect
+                 followRedirect=True, redirectLimit=20,
+                 afterFoundGet=False):
+        self.followRedirect = followRedirect
+        self.redirectLimit = redirectLimit
+        self._redirectCount = 0
         self.timeout = timeout
         self.agent = agent
-
+        self.afterFoundGet = afterFoundGet
         if cookies is None:
             cookies = {}
         self.cookies = cookies
@@ -241,6 +292,8 @@ class HTTPClientFactory(protocol.ClientFactory):
 
     def buildProtocol(self, addr):
         p = protocol.ClientFactory.buildProtocol(self, addr)
+        p.followRedirect = self.followRedirect
+        p.afterFoundGet = self.afterFoundGet
         if self.timeout:
             timeoutCall = reactor.callLater(self.timeout, p.timeout)
             self.deferred.addBoth(self._cancelTimeout, timeoutCall)
@@ -288,7 +341,9 @@ class HTTPDownloader(HTTPClientFactory):
 
     def __init__(self, url, fileOrName,
                  method='GET', postdata=None, headers=None,
-                 agent="Twisted client", supportPartial=0):
+                 agent="Twisted client", supportPartial=0,
+                 timeout=0, cookies=None, followRedirect=1,
+                 redirectLimit=20):
         self.requestedPartial = 0
         if isinstance(fileOrName, types.StringTypes):
             self.fileName = fileOrName
@@ -302,11 +357,14 @@ class HTTPDownloader(HTTPClientFactory):
                     headers["range"] = "bytes=%d-" % fileLength
         else:
             self.file = fileOrName
-        HTTPClientFactory.__init__(self, url, method=method, postdata=postdata, headers=headers, agent=agent)
-        self.deferred = defer.Deferred()
-        self.waiting = 1
+        HTTPClientFactory.__init__(
+            self, url, method=method, postdata=postdata, headers=headers,
+            agent=agent, timeout=timeout, cookies=cookies,
+            followRedirect=followRedirect, redirectLimit=redirectLimit)
+
 
     def gotHeaders(self, headers):
+        HTTPClientFactory.gotHeaders(self, headers)
         if self.requestedPartial:
             contentRange = headers.get("content-range", None)
             if not contentRange:
@@ -318,6 +376,7 @@ class HTTPDownloader(HTTPClientFactory):
                 # server is acting wierdly
                 self.requestedPartial = 0
 
+
     def openFile(self, partialContent):
         if partialContent:
             file = open(self.fileName, 'rb+')
@@ -334,7 +393,6 @@ class HTTPDownloader(HTTPClientFactory):
         if partialContent and not self.requestedPartial:
             raise ValueError, "we shouldn't get partial content response if we didn't want it!"
         if self.waiting:
-            self.waiting = 0
             try:
                 if not self.file:
                     self.file = self.openFile(partialContent)
@@ -352,7 +410,24 @@ class HTTPDownloader(HTTPClientFactory):
             self.file = None
             self.deferred.errback(failure.Failure())
 
+
+    def noPage(self, reason):
+        """
+        Close the storage file and errback the waiting L{Deferred} with the
+        given reason.
+        """
+        if self.waiting:
+            self.waiting = 0
+            if self.file:
+                try:
+                    self.file.close()
+                except:
+                    log.err(None, "Error closing HTTPDownloader file")
+            self.deferred.errback(reason)
+
+
     def pageEnd(self):
+        self.waiting = 0
         if not self.file:
             return
         try:
@@ -363,6 +438,7 @@ class HTTPDownloader(HTTPClientFactory):
         self.deferred.callback(self.value)
 
 
+
 def _parse(url, defaultPort=None):
     """
     Split the given URL into the scheme, host, port, and path.
@@ -380,55 +456,172 @@ def _parse(url, defaultPort=None):
     url = url.strip()
     parsed = http.urlparse(url)
     scheme = parsed[0]
-    path = urlunparse(('','')+parsed[2:])
+    path = urlunparse(('', '') + parsed[2:])
+
     if defaultPort is None:
         if scheme == 'https':
             defaultPort = 443
         else:
             defaultPort = 80
+
     host, port = parsed[1], defaultPort
     if ':' in host:
         host, port = host.split(':')
-        port = int(port)
-    if path == "":
-        path = "/"
+        try:
+            port = int(port)
+        except ValueError:
+            port = defaultPort
+
+    if path == '':
+        path = '/'
+
     return scheme, host, port, path
 
 
+def _makeGetterFactory(url, factoryFactory, contextFactory=None,
+                       *args, **kwargs):
+    """
+    Create and connect an HTTP page getting factory.
+
+    Any additional positional or keyword arguments are used when calling
+    C{factoryFactory}.
+
+    @param factoryFactory: Factory factory that is called with C{url}, C{args}
+        and C{kwargs} to produce the getter
+
+    @param contextFactory: Context factory to use when creating a secure
+        connection, defaulting to C{None}
+
+    @return: The factory created by C{factoryFactory}
+    """
+    scheme, host, port, path = _parse(url)
+    factory = factoryFactory(url, *args, **kwargs)
+    if scheme == 'https':
+        from twisted.internet import ssl
+        if contextFactory is None:
+            contextFactory = ssl.ClientContextFactory()
+        reactor.connectSSL(host, port, factory, contextFactory)
+    else:
+        reactor.connectTCP(host, port, factory)
+    return factory
+
+
 def getPage(url, contextFactory=None, *args, **kwargs):
-    """Download a web page as a string.
+    """
+    Download a web page as a string.
 
     Download a page. Return a deferred, which will callback with a
     page (as a string) or errback with a description of the error.
 
     See HTTPClientFactory to see what extra args can be passed.
     """
-    scheme, host, port, path = _parse(url)
-    factory = HTTPClientFactory(url, *args, **kwargs)
-    if scheme == 'https':
-        from twisted.internet import ssl
-        if contextFactory is None:
-            contextFactory = ssl.ClientContextFactory()
-        reactor.connectSSL(host, port, factory, contextFactory)
-    else:
-        reactor.connectTCP(host, port, factory)
-    return factory.deferred
+    return _makeGetterFactory(
+        url,
+        HTTPClientFactory,
+        contextFactory=contextFactory,
+        *args, **kwargs).deferred
 
 
 def downloadPage(url, file, contextFactory=None, *args, **kwargs):
-    """Download a web page to a file.
+    """
+    Download a web page to a file.
 
     @param file: path to file on filesystem, or file-like object.
 
     See HTTPDownloader to see what extra args can be passed.
     """
-    scheme, host, port, path = _parse(url)
-    factory = HTTPDownloader(url, file, *args, **kwargs)
-    if scheme == 'https':
-        from twisted.internet import ssl
-        if contextFactory is None:
-            contextFactory = ssl.ClientContextFactory()
-        reactor.connectSSL(host, port, factory, contextFactory)
-    else:
-        reactor.connectTCP(host, port, factory)
-    return factory.deferred
+    factoryFactory = lambda url, *a, **kw: HTTPDownloader(url, file, *a, **kw)
+    return _makeGetterFactory(
+        url,
+        factoryFactory,
+        contextFactory=contextFactory,
+        *args, **kwargs).deferred
+
+
+# The code which follows is based on the new HTTP client implementation.  It
+# should be significantly better than anything above, though it is not yet
+# feature equivalent.
+
+from twisted.internet.protocol import ClientCreator
+from twisted.web.error import SchemeNotSupported
+from twisted.web._newclient import ResponseDone, Request, HTTP11ClientProtocol
+from twisted.web._newclient import Response
+
+class Agent(object):
+    """
+    L{Agent} is a very basic HTTP client.  It supports I{HTTP} scheme URIs.  It
+    does not support persistent connections.
+
+    @ivar _reactor: The L{IReactorTCP} implementation which will be used to set
+        up connections over which to issue requests.
+
+    @since: 9.0
+    """
+    _protocol = HTTP11ClientProtocol
+
+    def __init__(self, reactor):
+        self._reactor = reactor
+
+
+    def request(self, method, uri, headers=None, bodyProducer=None):
+        """
+        Issue a new request.
+
+        @param method: The request method to send.
+        @type method: C{str}
+
+        @param uri: The request URI send.
+        @type uri: C{str}
+
+        @param headers: The request headers to send.  If no I{Host} header is
+            included, one will be added based on the request URI.
+        @type headers: L{Headers}
+
+        @param bodyProducer: An object which will produce the request body or,
+            if the request body is to be empty, L{None}.
+        @type bodyProducer: L{IBodyProducer} provider
+
+        @return: A L{Deferred} which fires with the result of the request (a
+            L{Response} instance), or fails if there is a problem setting up a
+            connection over which to issue the request.  It may also fail with
+            L{SchemeNotSupported} if the scheme of the given URI is not
+            supported.
+        @rtype: L{Deferred}
+        """
+        scheme, host, port, path = _parse(uri)
+        if scheme != 'http':
+            return defer.fail(SchemeNotSupported(
+                    "Unsupported scheme: %r" % (scheme,)))
+        cc = ClientCreator(self._reactor, self._protocol)
+        d = cc.connectTCP(host, port)
+        if headers is None:
+            headers = Headers()
+        if not headers.hasHeader('host'):
+            # This is a lot of copying.  It might be nice if there were a bit
+            # less.
+            headers = Headers(dict(headers.getAllRawHeaders()))
+            headers.addRawHeader(
+                'host', self._computeHostValue(scheme, host, port))
+        def cbConnected(proto):
+            return proto.request(Request(method, path, headers, bodyProducer))
+        d.addCallback(cbConnected)
+        return d
+
+
+    def _computeHostValue(self, scheme, host, port):
+        """
+        Compute the string to use for the value of the I{Host} header, based on
+        the given scheme, host name, and port number.
+        """
+        if port == 80:
+            return host
+        return '%s:%d' % (host, port)
+
+
+
+__all__ = [
+    'PartialDownloadError',
+    'HTTPPageGetter', 'HTTPPageDownloader', 'HTTPClientFactory', 'HTTPDownloader',
+    'getPage', 'downloadPage',
+
+    'ResponseDone', 'Response', 'Agent']
